package ca.dal.cs.schememm.parser;

import ca.dal.cs.schememm.common.CharStream;
import ca.dal.cs.schememm.common.Pair;
import ca.dal.cs.schememm.parser.builder.ParserBuilder;
import ca.dal.cs.schememm.parser.builder.Symbol;
import ca.dal.cs.schememm.parser.parsetree.Node;
import ca.dal.cs.schememm.parser.parsetree.NonTerminal;
import ca.dal.cs.schememm.parser.parsetree.Terminal;
import ca.dal.cs.schememm.scanner.LexicalError;
import ca.dal.cs.schememm.scanner.Scanner;
import ca.dal.cs.schememm.scanner.Token;

import java.io.IOException;
import java.util.*;

public class Parser {

    /**
     * Main function to run the parser
     */
    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println("USAGE: java Parser <scheme-- file>");
            System.exit(1);
        }
        try {
            CharStream chars     = CharStream.fromFile(args[0]);
            Scanner    tokens    = new Scanner(chars);
            Parser     parser    = ParserBuilder.run();
            Node       parseTree = parser.run(tokens);
            parseTree.format("");
        }
        catch (Exception e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
    }

    /**
     * Constructor to build a parser
     * @param startSymbol the start symbol to initialize the parser with
     * @param expected the set of terminals that can start the expansion for
     *                 a given terminal (used for error reporting)
     * @param productions the lookup table of productions in the grammar to
     *                    be used by the parser.
     */
    public Parser(String startSymbol,
                  Map<String, Set<Token.Name>> expected,
                  Map<Pair<String, Token.Name>, List<Symbol>> productions) {
        this.startSymbol = startSymbol;
        this.expected = expected;
        this.productions = productions;
    }

    /**
     * Run this parser on a given token stream
     * @param tokens the token stream to be parsed
     * @return the root of the computed parse tree
     */
    private Node run(Scanner tokens)
            throws IOException, LexicalError, SyntaxError {
        Node        root           = new NonTerminal(null, startSymbol);
        Stack<Node> sententialForm = new Stack<>();
        sententialForm.push(root);
        while (!sententialForm.empty()) {
            Node        front  = sententialForm.pop();
            NonTerminal expand = front.asNonTerminal();
            if (expand == null) {
                Terminal expected = front.asTerminal();
                if (tokens.item().id() == expected.name()) {
                    expected.setString(tokens.item().string());
                    tokens.step();
                }
                else {
                    throw new SyntaxError(tokens.item(), tokens.pos(),
                                          Collections.singleton(
                                                  expected.name()));
                }
            }
            else {
                List<Symbol> production =
                        productions.get(new Pair<>(expand.name(),
                                                   tokens.item().id()));
                if (production == null) {
                    throw new SyntaxError(tokens.item(), tokens.pos(),
                                          expected.get(expand.name()));
                }
                else {
                    for (Symbol sym : production) {
                        sententialForm.push(sym.node(expand));
                    }
                }
            }
        }
        if (tokens.item().id() != Token.Name.EOF) {
            throw new SyntaxError(tokens.item(),
                                  tokens.pos(),
                                  Collections.singleton(null));
        }
        return root;
    }

    private String                                      startSymbol;
    private Map<String, Set<Token.Name>>                expected;
    private Map<Pair<String, Token.Name>, List<Symbol>> productions;
}
