package ca.dal.cs.schememm.parser.builder;

import ca.dal.cs.schememm.parser.ParserError;
import ca.dal.cs.schememm.parser.parsetree.Node;
import ca.dal.cs.schememm.scanner.Token;

import java.util.*;

/**
 * An abstract base class for symbols in a grammar.  Concrete subclasses are
 * Terminal and NonTerminal.
 */
public abstract class Symbol {

    /**
     * Construct a new symbol (terminal or non-terminal) from its name.  The
     * type is determined by the name's prefix (T. or V.).
     *
     * @param name the name of the symbol to be created
     * @return the created symbol
     */
    static Symbol symbol(String name) throws ParserError {
        if (name.startsWith("T.")) {
            return new Terminal(name);
        }
        else if (name.startsWith("V.")) {
            return new NonTerminal(name);
        }
        else {
            throw new ParserError("Grammar symbols must start with T. or V.");
        }

    }

    /**
     * Create a new symbol
     */
    Symbol() {
        firstSet = new HashSet<>();
    }

    /**
     * Return a const reference to this symbol's first set
     *
     * @return the first set
     */
    Set<Token.Name> firstSet() {
        return Collections.unmodifiableSet(firstSet);
    }

    /**
     * Add more terminals to this symbol's first set
     *
     * @param additions a list of character ranges to be added
     * @return true if the set was changed
     */
    boolean augmentFirstSet(Set<Token.Name> additions) {
        return firstSet.addAll(additions);
    }

    /**
     * Cast this symbol to a non-terminal.  This produces null if the object is
     * not a non-terminal.
     *
     * @return the symbol cast to non-terminal
     */
    public NonTerminal asNonTerminal() {
        return null;
    }

    /**
     * Construct a parse tree node corresponding to this symbol
     *
     * @param parent the parent of the parse tree node
     * @return the created node
     */
    public abstract Node node(
            ca.dal.cs.schememm.parser.parsetree.NonTerminal parent);

    /**
     * The name of this symbol
     */
    public abstract String name();

    private Set<Token.Name> firstSet;
}
