package ca.dal.cs.schememm.scanner.fa;

import apple.laf.JRSUIState;
import ca.dal.cs.schememm.common.Pair;
import ca.dal.cs.schememm.scanner.Token;
import ca.dal.cs.schememm.scanner.regex.RegexError;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import static org.testng.Assert.*;

public class NFATest {

    @Test
    public void testEpsilonNFA() {
        NFA        nfa    = NFA.epsilonNFA(Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 1);
        assertEquals(nfa.start.token, Token.Name.NUMBER);
        assertSame(nfa.accepting.get(0), nfa.start);
    }

    @Test
    public void testCharRangeNFA() {
        NFA nfa =
                NFA.charRangeNFA(new Pair<>(97, 110), Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 2);
        assertNull(nfa.start.token);
        assertSame(nfa.accepting.get(0).token, Token.Name.NUMBER);
    }

    @Test
    public void testParallelJoin() {
        NFA nfa1 =
                NFA.charRangeNFA(new Pair<>(97, 110), Token.Name.NUMBER);
        NFA        nfa2   = NFA.epsilonNFA(Token.Name.IDENTIFIER);
        NFA        nfa    = NFA.parallelJoin(Arrays.asList(nfa1, nfa2));
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 4);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.size(), 2);
        assertTrue(nfa.accepting.containsAll(
                Arrays.asList(nfa1.accepting.get(0), nfa2.accepting.get(0))));
    }

    @Test
    public void testSequentialJoin() {
        NFA nfa1 =
                NFA.charRangeNFA(new Pair<>(97, 110), Token.Name.NUMBER);
        NFA        nfa2   = NFA.epsilonNFA(Token.Name.IDENTIFIER);
        NFA        nfa    = NFA.sequentialJoin(Arrays.asList(nfa1, nfa2));
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 3);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.size(), 1);
        assertTrue(nfa.accepting.contains(nfa2.accepting.get(0)));
    }

    @Test
    public void testKleeneStar() {
        NFA nfa1 =
                NFA.charRangeNFA(new Pair<>(97, 110), Token.Name.NUMBER);
        NFA        nfa2   = NFA.epsilonNFA(Token.Name.IDENTIFIER);
        NFA        nfa3   = NFA.parallelJoin(Arrays.asList(nfa1, nfa2));
        NFA        nfa    = NFA.kleeneStar(nfa3, Token.Name.STRING);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 4);
        assertEquals(nfa.start.token, Token.Name.STRING);
        assertEquals(nfa.accepting.size(), 1);
        assertTrue(nfa.accepting.contains(nfa.start));
    }

    @Test
    public void testFromEpsilonRegex() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 1);
        assertEquals(nfa.start.token, Token.Name.NUMBER);
        assertSame(nfa.accepting.get(0), nfa.start);
    }

    @Test
    public void testFromCharRangeRegex1() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("9", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 2);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.get(0).token, Token.Name.NUMBER);
        assertEquals(nfa.start.neighbours.get(0).left,
                     new Pair<>((int) '9', (int) '9'));
    }

    @Test
    public void testFromCharRangeRegex2() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("[a-z]", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 2);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.get(0).token, Token.Name.NUMBER);
        assertEquals(nfa.start.neighbours.get(0).left,
                     new Pair<>((int) 'a', (int) 'z'));
    }

    @Test
    public void testFromParallelJoinRegex1() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("9|[a-z]", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 5);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.size(), 2);
        assertEquals(nfa.accepting.get(0).token, Token.Name.NUMBER);
        assertEquals(nfa.accepting.get(1).token, Token.Name.NUMBER);
    }

    @Test
    public void testFromParallelJoinRegex2() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("[a-z9]", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 5);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.size(), 2);
        assertEquals(nfa.accepting.get(0).token, Token.Name.NUMBER);
        assertEquals(nfa.accepting.get(1).token, Token.Name.NUMBER);
    }

    @Test
    public void testFromSequentialJoinRegex() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("x[a-z9]", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 7);
        assertNull(nfa.start.token);
        assertEquals(nfa.accepting.size(), 2);
        assertEquals(nfa.accepting.get(0).token, Token.Name.NUMBER);
        assertEquals(nfa.accepting.get(1).token, Token.Name.NUMBER);
    }

    @Test
    public void testFromKleeneStarRegex() throws IOException, RegexError {
        NFA        nfa    = NFA.fromRegex("(x[a-z9])*", Token.Name.NUMBER);
        Set<State> states = new HashSet<>();
        State.collectStates(nfa.start, states);

        assertEquals(states.size(), 7);
        assertEquals(nfa.start.token, Token.Name.NUMBER);
        assertEquals(nfa.accepting.size(), 1);
        assertSame(nfa.accepting.get(0), nfa.start);
    }

    @Test
    public void testFromComplexRegex1() throws IOException, RegexError {
        NFA nfa = NFA.fromRegex("a?b+(c|d){1,2}", Token.Name.NUMBER);
        State[] s = new State[20];
        Pair<Pair<Integer, Integer>, State> nbr1, nbr2;

        s[0] = nfa.start;

        assertEquals(s[0].neighbours.size(), 2);
        nbr1 = s[0].neighbours.get(0);
        nbr2 = s[0].neighbours.get(1);

        assertNull(nbr1.left);
        assertNull(nbr2.left);
        s[1] = nbr1.right;
        s[3] = nbr2.right;

        assertEquals(s[1].neighbours.size(), 1);
        nbr1 = s[1].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'a', (int) 'a'));
        s[2] = nbr1.right;

        assertEquals(s[2].neighbours.size(), 1);
        nbr1 = s[2].neighbours.get(0);

        assertNull(nbr1.left);
        s[4] = nbr1.right;

        assertEquals(s[3].neighbours.size(), 1);
        nbr1 = s[3].neighbours.get(0);

        assertNull(nbr1.left);
        assertSame(nbr1.right, s[4]);

        assertEquals(s[4].neighbours.size(), 1);
        nbr1 = s[4].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'b', (int) 'b'));
        s[5] = nbr1.right;

        assertEquals(s[5].neighbours.size(), 1);
        nbr1 = s[5].neighbours.get(0);

        assertNull(nbr1.left);
        s[6] = nbr1.right;

        assertEquals(s[6].neighbours.size(), 2);
        nbr1 = s[6].neighbours.get(0);
        nbr2 = s[6].neighbours.get(1);

        assertEquals(nbr1.left, new Pair<>((int) 'b', (int) 'b'));
        assertNull(nbr2.left);
        s[7] = nbr1.right;
        s[8] = nbr2.right;

        assertEquals(s[7].neighbours.size(), 1);
        nbr1 = s[7].neighbours.get(0);

        assertNull(nbr1.left);
        assertSame(nbr1.right, s[6]);

        assertEquals(s[8].neighbours.size(), 2);
        nbr1 = s[8].neighbours.get(0);
        nbr2 = s[8].neighbours.get(1);

        assertNull(nbr1.left);
        assertNull(nbr2.left);
        s[9] = nbr1.right;
        s[11] = nbr2.right;

        assertEquals(s[9].neighbours.size(), 1);
        nbr1 = s[9].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'c', (int) 'c'));
        s[10] = nbr1.right;

        assertEquals(s[10].neighbours.size(), 1);
        nbr1 = s[10].neighbours.get(0);

        assertNull(nbr1.left);
        s[13] = nbr1.right;

        assertEquals(s[11].neighbours.size(), 1);
        nbr1 = s[11].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'd', (int) 'd'));
        s[12] = nbr1.right;

        assertEquals(s[12].neighbours.size(), 1);
        nbr1 = s[12].neighbours.get(0);

        assertNull(nbr1.left);
        assertSame(nbr1.right, s[13]);

        assertEquals(s[13].neighbours.size(), 2);
        nbr1 = s[13].neighbours.get(0);
        nbr2 = s[13].neighbours.get(1);

        assertNull(nbr1.left);
        assertNull(nbr2.left);
        s[14] = nbr1.right;
        s[19] = nbr2.right;

        assertEquals(s[14].neighbours.size(), 2);
        nbr1 = s[14].neighbours.get(0);
        nbr2 = s[14].neighbours.get(1);

        assertNull(nbr1.left);
        assertNull(nbr2.left);
        s[15] = nbr1.right;
        s[17] = nbr2.right;

        assertEquals(s[15].neighbours.size(), 1);
        nbr1 = s[15].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'c', (int) 'c'));
        s[16] = nbr1.right;

        assertEquals(s[16].neighbours.size(), 0);

        assertEquals(s[17].neighbours.size(), 1);
        nbr1 = s[17].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'd', (int) 'd'));
        s[18] = nbr1.right;

        assertEquals(s[18].neighbours.size(), 0);

        assertEquals(s[19].neighbours.size(), 0);

        for (int i = 0; i < 20; ++i) {
            if (i == 16 || i == 18 || i == 19) {
                assertEquals(s[i].token, Token.Name.NUMBER);
            } else {
                assertNull(s[i].token);
            }
        }
    }

    @Test
    public void testFromComplexRegex2() throws IOException, RegexError {
        NFA nfa = NFA.fromRegex("[a-z0-9]*[^a-z]", Token.Name.IDENTIFIER);
        State[] s = new State[10];
        Pair<Pair<Integer, Integer>, State> nbr1, nbr2, nbr3;

        s[0] = nfa.start;

        assertEquals(s[0].neighbours.size(), 3);
        nbr1 = s[0].neighbours.get(0);
        nbr2 = s[0].neighbours.get(1);
        nbr3 = s[0].neighbours.get(2);

        assertNull(nbr1.left);
        assertNull(nbr2.left);
        assertNull(nbr3.left);
        s[1] = nbr1.right;
        s[3] = nbr2.right;
        s[5] = nbr3.right;

        assertEquals(s[1].neighbours.size(), 1);
        nbr1 = s[1].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'a', (int) 'z'));
        s[2] = nbr1.right;

        assertEquals(s[2].neighbours.size(), 1);
        nbr1 = s[2].neighbours.get(0);

        assertNull(nbr1.left);
        assertSame(nbr1.right, s[0]);

        assertEquals(s[3].neighbours.size(), 1);
        nbr1 = s[3].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) '0', (int) '9'));
        s[4] = nbr1.right;

        assertEquals(s[4].neighbours.size(), 1);
        nbr1 = s[4].neighbours.get(0);

        assertNull(nbr1.left);
        assertSame(nbr1.right, s[0]);

        assertEquals(s[5].neighbours.size(), 2);
        nbr1 = s[5].neighbours.get(0);
        nbr2 = s[5].neighbours.get(1);

        assertNull(nbr1.left);
        assertNull(nbr2.left);
        s[6] = nbr1.right;
        s[8] = nbr2.right;

        assertEquals(s[6].neighbours.size(), 1);
        nbr1 = s[6].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>(0, (int) 'a' - 1));
        s[7] = nbr1.right;

        assertEquals(s[7].neighbours.size(), 0);

        assertEquals(s[8].neighbours.size(), 1);
        nbr1 = s[8].neighbours.get(0);

        assertEquals(nbr1.left, new Pair<>((int) 'z' + 1,
                                           (int) Character.MAX_VALUE));
        s[9] = nbr1.right;

        assertEquals(s[9].neighbours.size(), 0);

        for (int i = 0; i < 10; ++i) {
            if (i == 7 || i == 9) {
                assertEquals(s[i].token, Token.Name.IDENTIFIER);
            } else {
                assertNull(s[i].token);
            }
        }
    }
}
